from __future__ import annotations

import re

import roman

from ._morph import get_morph
from .constants import KNOWN_ABBREVIATIONS
from .numerals._helpers import get_numeral_case, simple_tokenize
from .options import NormalizeOptions

_ROMAN_ABBREVIATION_EXCEPTIONS = frozenset(
    {
        *(abbr for abbr in KNOWN_ABBREVIATIONS if re.fullmatch(r"[A-Z-]+", abbr)),
        "DC",
        "MD",
        "CV",
        "CI",
        "DI",
        "MC",
        "CM",
        "MM",
    }
)

_CENTURY_CASE_TO_ENDING = {
    "nomn": "-й",
    "accs": "-й",
    "gent": "-го",
    "datv": "-му",
    "ablt": "-м",
    "loct": "-м",
}


def _expand_century_abbreviation(text: str, match: re.Match[str], number: int) -> str:
    morph = get_morph()
    left_context = text[max(0, match.start() - 60) : match.start()]
    right_context = text[match.end() : match.end() + 60]
    tokens = simple_tokenize(left_context) + [str(number), "век"] + simple_tokenize(right_context)
    case = get_numeral_case(tokens, len(simple_tokenize(left_context)))
    ending = _CENTURY_CASE_TO_ENDING.get(case, "-й")
    parsed = morph.parse("век")[0]
    noun_form = parsed.inflect({case, "sing"})
    century_word = noun_form.word if noun_form else "век"
    return f"{number}{ending} {century_word}"


def convert_roman_words(text: str) -> str:
    words = {
        "в.": ("век", "-й"),
        "век": ("век", "-й"),
        "века": ("века", "-го"),
        "веку": ("веку", "-му"),
        "веке": ("веке", "-м"),
        "веком": ("веком", "-м"),
        "веками": ("веками", "-ми"),
        "веках": ("веках", "-х"),
        "столетие": ("столетие", "-е"),
        "столетия": ("столетия", "-го"),
        "столетию": ("столетию", "-му"),
        "столетии": ("столетии", "-м"),
        "столетием": ("столетием", "-м"),
        "столетиях": ("столетиях", "-х"),
        "глава": ("глава", "-я"),
        "главы": ("главы", "-й"),
        "главе": ("главе", "-й"),
        "главу": ("главу", "-ю"),
        "главой": ("главой", "-й"),
        "главами": ("главами", "-ми"),
        "главах": ("главах", "-х"),
        "раздел": ("раздел", "-й"),
        "раздела": ("раздела", "-го"),
        "разделе": ("разделе", "-м"),
        "разделу": ("разделу", "-му"),
        "часть": ("часть", "-я"),
        "части": ("части", "-й"),
        "частью": ("частью", "-й"),
        "том": ("том", "-й"),
        "тома": ("тома", "-го"),
        "томе": ("томе", "-м"),
        "томом": ("томом", "-м"),
        "книга": ("книга", "-я"),
        "книги": ("книги", "-й"),
        "книге": ("книге", "-й"),
        "книгой": ("книгой", "-й"),
        "квартал": ("квартал", "-й"),
        "квартала": ("квартала", "-го"),
        "кварталу": ("кварталу", "-му"),
        "квартале": ("квартале", "-м"),
        "кварталом": ("кварталом", "-м"),
        "кв.": ("квартал", "-й"),
        "кв": ("квартал", "-й"),
        "в": ("век", "-й"),
    }
    suffixes_regex = "|".join(map(re.escape, words.keys()))
    pattern = rf"\b([IVXLCDM]+)\s*({suffixes_regex})(?!\w)"

    def repl(match: re.Match[str]) -> str:
        try:
            number = roman.fromRoman(match.group(1).upper())
        except roman.InvalidRomanNumeralError:
            return match.group(0)
        matched_suffix = match.group(2).lower()
        if matched_suffix in {"в.", "в"}:
            return _expand_century_abbreviation(text, match, number)
        target_word, ending = words[matched_suffix]
        return f"{number}{ending} {target_word}"

    return re.sub(pattern, repl, text)


def convert_roman_names(text: str) -> str:
    pattern = r"\b([А-ЯЁ][а-яё]+)\s+(II|III|IV|V|VI|VII|VIII|IX|X|XI|XII|XIII|XIV|XV)\b"

    def repl(match: re.Match[str]) -> str:
        try:
            return f"{match.group(1)} {roman.fromRoman(match.group(2))}"
        except roman.InvalidRomanNumeralError:
            return match.group(0)

    return re.sub(pattern, repl, text)


def convert_heading_roman_numerals(text: str) -> str:
    heading_words = (
        "глава",
        "главы",
        "часть",
        "части",
        "раздел",
        "раздела",
        "разделе",
        "том",
        "тома",
        "томе",
        "книга",
        "книги",
        "книге",
    )
    pattern = rf"\b({'|'.join(map(re.escape, heading_words))})\s+([IVXLCDM]+)\b"

    def repl(match: re.Match[str]) -> str:
        if match.group(2) != match.group(2).upper():
            return match.group(0)
        try:
            return f"{match.group(1)} {roman.fromRoman(match.group(2).upper())}"
        except roman.InvalidRomanNumeralError:
            return match.group(0)

    return re.sub(pattern, repl, text, flags=re.IGNORECASE)


def convert_other_roman_numerals(text: str) -> str:
    pattern = r"\b([IVXLCDM]{2,})\b"

    def repl(match: re.Match[str]) -> str:
        token = match.group(1).upper()
        if token in _ROMAN_ABBREVIATION_EXCEPTIONS:
            return match.group(0)
        try:
            return str(roman.fromRoman(token))
        except roman.InvalidRomanNumeralError:
            return match.group(0)

    return re.sub(pattern, repl, text)


def normalize_cyrillic_roman(text: str) -> str:
    cyrillic_to_latin = {
        "Х": "X",
        "х": "X",
        "С": "C",
        "с": "C",
        "І": "I",
        "l": "I",
        "і": "I",
        "М": "M",
        "м": "M",
    }
    pattern = re.compile(r"\b([IVXLCDMХхСсІіМм]{2,})\b")

    def repl(match: re.Match[str]) -> str:
        word = match.group(1)
        if word == word.lower():
            return word
        if not any(char in "ХхСсІіМм" for char in word):
            return word
        latin_word = "".join(cyrillic_to_latin.get(char, char) for char in word).upper()
        try:
            roman.fromRoman(latin_word)
            return latin_word
        except roman.InvalidRomanNumeralError:
            return word

    return pattern.sub(repl, text)


def normalize_roman(text: str, options: NormalizeOptions | None = None) -> str:
    active = options or NormalizeOptions()
    if not active.enable_roman_normalization:
        return text
    text = normalize_cyrillic_roman(text)
    text = convert_roman_words(text)
    text = convert_roman_names(text)
    text = convert_heading_roman_numerals(text)
    text = convert_other_roman_numerals(text)
    return text



from __future__ import annotations

import re

import num2words

from .._morph import get_morph
from ._constants import HYPHENATED_WORD_PATTERN, ORDINAL_PATTERN
from ._helpers import get_numeral_case, inflect_numeral_string, simple_tokenize


def normalize_hyphenated_words(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        morph = get_morph()
        num_str = match.group(1)
        word = match.group(2)
        word_lower = word.lower()
        ordinal_suffixes = {"й", "я", "е", "го", "му", "м", "ю", "ее", "ий", "ая", "ое"}
        cardinal_case_suffixes = {"ти", "ми", "х", "мя", "и"}
        if word_lower in ordinal_suffixes:
            return match.group(0)
        if (
            word_lower
            in {
                "ый",
                "ой",
                "й",
                "ого",
                "го",
                "ому",
                "ым",
                "ом",
                "му",
                "е",
                "х",
                "м",
                "ми",
            }
            and int(num_str) > 100
        ):
            return match.group(0)
        try:
            int(num_str)
        except Exception:
            return match.group(0)
        case_from_suffix = None
        if word_lower == "х":
            case_from_suffix = None
        elif word_lower in ("ти", "и"):
            case_from_suffix = "gent"
        elif word_lower in ("ми", "мя"):
            case_from_suffix = "ablt"
        ctx_left = text[max(0, match.start() - 60) : match.start()]
        ctx_right = text[match.end() : match.end() + 60]
        tokens_left = simple_tokenize(ctx_left)
        tokens_right = simple_tokenize(ctx_right)
        context_case = get_numeral_case(
            tokens_left + [num_str] + tokens_right, len(tokens_left)
        )
        if case_from_suffix:
            case = case_from_suffix
        elif word_lower == "х":
            case = context_case if context_case in ("gent", "loct") else "gent"
        else:
            case = context_case
        p_word = morph.parse(word_lower)[0]
        is_adj_like = "ADJF" in p_word.tag or word_lower.endswith(
            (
                "дневный",
                "часовой",
                "минутный",
                "летний",
                "этажный",
                "тонный",
                "процентный",
                "кратный",
                "кратного",
                "кратном",
                "кратных",
            )
        )
        target_case = "gent" if is_adj_like and case == "nomn" else case
        num_words = inflect_numeral_string(num_str, target_case)
        if word_lower in cardinal_case_suffixes:
            return num_words
        return (
            f"{num_words}{word}"
            if is_adj_like
            else (f"{num_words}" if len(word) <= 3 else f"{num_words} {word}")
        )

    return HYPHENATED_WORD_PATTERN.sub(repl, text)


def normalize_ordinals(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        morph = get_morph()
        num_str = match.group(1)
        suffix = match.group(2).lower()
        try:
            num = int(num_str)
        except ValueError:
            return match.group(0)
        gender = "masc"
        is_cardinal_suffix = suffix in ("ти", "ми")
        ctx_left = text[max(0, match.start() - 60) : match.start()]
        ctx_right = text[match.end() : match.end() + 60]
        tokens_left = simple_tokenize(ctx_left)
        tokens_right = simple_tokenize(ctx_right)
        case = get_numeral_case(
            tokens_left + [num_str] + tokens_right, len(tokens_left)
        )
        if suffix in ("я", "яя"):
            gender = "femn"
        elif suffix in ("е", "ее"):
            gender = "neut"
        elif suffix == "й" and tokens_right:
            p_next = morph.parse(tokens_right[0].strip(".,!?;:"))[0]
            if "femn" in p_next.tag:
                gender = "femn"
                if case == "nomn":
                    case = "gent"
        case_from_suffix = None
        if suffix == "го":
            case_from_suffix = "gent"
        elif suffix == "му":
            case_from_suffix = "datv"
        elif suffix == "й" and case == "nomn" and tokens_right:
            p_next = morph.parse(tokens_right[0].strip(".,!?;:"))[0]
            if "femn" in p_next.tag:
                gender = "femn"
                case_from_suffix = "gent"
        if suffix == "м":
            if tokens_right:
                p_next = morph.parse(tokens_right[0].strip(".,!?;:"))[0]
                case_from_suffix = "loct" if "sing" in p_next.tag else "datv"
            else:
                case_from_suffix = "loct"
        case = case_from_suffix or case
        plural = suffix in ("х", "ми", "е", "м") and not (
            suffix == "м" and case == "loct"
        )
        cases_map = {
            "nomn": "nominative",
            "gent": "genitive",
            "datv": "dative",
            "accs": "accusative",
            "ablt": "instrumental",
            "loct": "prepositional",
        }
        gender_map = {"masc": "m", "femn": "f", "neut": "n"}
        if case in cases_map:
            try:
                kwargs = {"case": cases_map[case]}
                if not is_cardinal_suffix:
                    kwargs["to"] = "ordinal"
                if plural:
                    kwargs["plural"] = True
                elif gender in gender_map and not is_cardinal_suffix:
                    kwargs["gender"] = gender_map[gender]
                return num2words.num2words(num, lang="ru", **kwargs) + " "
            except Exception:
                pass
        try:
            ordinal = num2words.num2words(
                num, lang="ru", to="cardinal" if is_cardinal_suffix else "ordinal"
            )
        except Exception:
            return match.group(0)
        words = ordinal.split()
        if not words:
            return ordinal
        parsed = morph.parse(words[-1])
        if parsed:
            p = parsed[0]
            target_tags = {case}
            if plural:
                target_tags.add("plur")
            elif gender:
                target_tags.add(gender)
            inflected = p.inflect(target_tags)
            if inflected:
                words[-1] = inflected.word
        return " ".join(words) + " "

    return ORDINAL_PATTERN.sub(repl, text)



from __future__ import annotations

import re

from ._morph import get_morph
from .abbreviation_rules import ABBREVIATION_PATTERNS
from .constants import (
    ABBREV_DOTTED_PATTERN,
    ABBREV_MAX_LEN_EN,
    ABBREV_MAX_LEN_RU,
    ABBREV_MIN_LEN,
    ABBREV_SOLID_PATTERN,
    EN_LETTER_NAMES,
    KNOWN_ABBREVIATIONS,
    PERSON_INITIALS_SURNAME_PATTERN,
    PERSON_SURNAME_INITIALS_PATTERN,
    REAL_WORD_POS,
    RU_LETTER_NAMES,
    RU_VOWELS,
    RUSSIAN_NAME_TOKEN,
)
from .options import NormalizeOptions

_ETC_ABBREVIATION_PATTERN = re.compile(
    r"(?<!\w)(?P<abbr>т\.?\s*[дп]\.)(?P<tail>\s*)(?!\w)", re.IGNORECASE
)


def _expand_contextual_etc_abbreviations(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        raw_abbr = match.group("abbr")
        tail = match.group("tail")
        normalized_abbr = re.sub(r"\s+", "", raw_abbr).lower()
        expansion = "так далее" if "д" in normalized_abbr else "тому подобное"

        rest = text[match.end() :]
        next_char = rest[:1]
        stripped_rest = rest.lstrip()
        next_significant = stripped_rest[:1]
        has_linebreak_before_next = "\n" in tail or (
            "\n" in rest and rest[: len(rest) - len(stripped_rest)] != ""
        )

        keep_terminal_dot = (
            not stripped_rest
            or has_linebreak_before_next
            or (next_significant and next_significant.isupper())
        )

        if (
            next_char in '"»”)]}'
            and len(stripped_rest) > 1
            and stripped_rest[1:2].isupper()
        ):
            keep_terminal_dot = True

        return f"{expansion}{'.' if keep_terminal_dot else ''}{tail}"

    return _ETC_ABBREVIATION_PATTERN.sub(repl, text)


def expand_person_initials(text: str) -> str:
    def initial_name(ch: str) -> str:
        return RU_LETTER_NAMES.get(ch.upper(), ch.lower())

    def choose_tail(match_end: int) -> tuple[bool, str]:
        tail = text[match_end:]
        stripped = tail.lstrip()
        if not stripped:
            return True, ""
        if stripped[0] in ".!?…":
            return True, ""
        if stripped[0] in ",;:":
            return False, ""

        if re.match(rf"^\s+{RUSSIAN_NAME_TOKEN}\s+[А-ЯЁ]\.\s*[А-ЯЁ]\.", tail):
            return False, ","
        if stripped[0].islower() or stripped[0].isdigit():
            return False, ","
        return True, ""

    def repl_surname_initials(match):
        surname = match.group("surname")
        i1 = initial_name(match.group("i1"))
        i2 = initial_name(match.group("i2"))
        is_final, non_final_tail = choose_tail(match.end())
        if is_final:
            terminal = (
                ""
                if text[match.end() :].lstrip().startswith((".", "!", "?", "…"))
                else "."
            )
            return f"{surname} {i1} {i2}{terminal}"
        return f"{surname}, {i1}, {i2}{non_final_tail}"

    def repl_initials_surname(match):
        i1 = initial_name(match.group("i1"))
        i2 = initial_name(match.group("i2"))
        surname = match.group("surname")
        is_final, non_final_tail = choose_tail(match.end())
        if is_final:
            terminal = (
                ""
                if text[match.end() :].lstrip().startswith((".", "!", "?", "…"))
                else "."
            )
            return f"{i1} {i2} {surname}{terminal}"
        return f"{i1}, {i2}, {surname}{non_final_tail}"

    text = PERSON_SURNAME_INITIALS_PATTERN.sub(repl_surname_initials, text)
    return PERSON_INITIALS_SURNAME_PATTERN.sub(repl_initials_surname, text)


def expand_letter_abbreviations(text: str) -> str:
    def expand_token(token: str, is_dotted: bool) -> str:
        tail_dot = token.endswith(".")
        letters = [char for char in token if char.isalpha()]
        if len(letters) < ABBREV_MIN_LEN:
            return token

        letters_upper = [char.upper() for char in letters]
        joined_key = "".join(letters_upper)
        if joined_key in KNOWN_ABBREVIATIONS:
            custom_reading = KNOWN_ABBREVIATIONS[joined_key]
            return (
                custom_reading + "."
                if custom_reading and tail_dot
                else custom_reading or token
            )

        if all(char in RU_LETTER_NAMES for char in letters_upper):
            if not is_dotted:
                parsed = get_morph().parse(joined_key.lower())[0]
                if parsed.tag.POS in REAL_WORD_POS and "Abbr" not in parsed.tag:
                    return token
            if not is_dotted:
                if len(letters_upper) > ABBREV_MAX_LEN_RU:
                    return token
                if sum(1 for char in letters_upper if char in RU_VOWELS) > 1:
                    return token
            names = [RU_LETTER_NAMES[char] for char in letters_upper]
        elif all(char in EN_LETTER_NAMES for char in letters_upper):
            if not is_dotted and len(letters_upper) > ABBREV_MAX_LEN_EN:
                return token
            names = [EN_LETTER_NAMES[char] for char in letters_upper]
        else:
            return token

        expanded = " ".join(names)
        return expanded + "." if tail_dot else expanded

    text = ABBREV_DOTTED_PATTERN.sub(lambda m: expand_token(m.group(0), True), text)
    return ABBREV_SOLID_PATTERN.sub(lambda m: expand_token(m.group(0), False), text)


def expand_abbreviations(text: str, options: NormalizeOptions | None = None) -> str:
    active = options or NormalizeOptions()
    if not active.enable_abbreviation_expansion:
        return text
    if active.enable_contextual_abbreviation_expansion:
        text = _expand_contextual_etc_abbreviations(text)
        for pattern, replacement in ABBREVIATION_PATTERNS:
            text = pattern.sub(replacement, text)
    if active.enable_initials_expansion:
        text = expand_person_initials(text)
    if active.enable_letter_abbreviation_expansion:
        text = expand_letter_abbreviations(text)
    return text
